!=========================================================================
!      MODULE FOR BICONJUGATE GRADIENT SOLVER WITH PRECONDITIONING
!
! This module is inspired by the book "Numerical Recipes", but programmed
! by C.P. Dullemond, May 2010.
!=========================================================================
module bicg_module

contains

!-------------------------------------------------------------------------
!                     BICONJUGATE GRADIENT SOLVER
!
! This subroutine solves a set of coupled linear equations, written in 
! matrix form as:
!
!     M # sol = rhs
!
! Where # is the matrix multiplication of matrix M with vector sol. The
! right-hand-side vector rhs is given as a 1D array, and the matrix M is 
! given in terms of a set of user-specified subroutines that perform
! the multiplication of M or Transpose(M) with any vector. This allows
! the user to 100% decide for him/herself how to store the matrix.
! In addition, the user must give also subroutines that solve approximate
! versions of the above equation:
!
!     A # vec1 = vec2      (where vec2 will be given to you by the program)
!
! where A is an approximation of matrix M to be 100% decided by the user,
! but chosen such that (a) the above equation is easily solved for vec1,
! given any vec2 and (b) is as close as possible to matrix M. Matrix A
! is the preconditioning-matrix. If such a choice cannot be easily found,
! then A=unit-matrix is a good choice, and A=Diag(M)^inv is a better 
! choice (the latter being the Jacobi approximate operator). But if an
! even better choice can be found, then convergence might be even better.
!
! ARGUMENTS:
!  n                  Nr of components of the vector
!  rhs                Vector on the right-hand-side
!  sol                Initial guess for the solution
!  mx                 User-provided subroutine: Matrix*Vector
!  mtrx               User-provided subroutine: Transpose(Matrix)*Vector
!  msol               User-provided subroutine: Approximate solution
!  mtrsol             User-provided subroutine: Approximate solution with transpose
!  itol               Which kind of error estimate do we use?
!                     We use roughly the itol versions of Num Recip
!
! RETURNS:
!  sol                The solution vector
!-------------------------------------------------------------------------
subroutine solve_bicg(n,rhs,mx,mtrx,msol,mtrsol,itermax,inorm,itol,tol,sol)
  implicit none
  integer :: n,itol,itermax,inorm
  double precision :: rhs(n),sol(n),tol
  external :: mx,mtrx,msol,mtrsol
  double precision, allocatable :: r(:),rb(:),p(:),pb(:),vec(:),z(:),zb(:)
  integer :: iter
  double precision :: val1,val2,alpha,beta,eqnorm,err
  !
  ! Allocate vectors r, rb, p, pb
  !
  allocate(r(n),rb(n),p(n),pb(n),z(n),zb(n),vec(n))
  !
  ! Make an initial value for r(:) and rb(:), based on the initial guess
  ! for the solution sol(:)
  !
  call mx(n,sol,r)
  r  = rhs - r
  rb = r
  !
  ! Compute the norm of the equation
  !
  if(itol.eq.0) then
     !
     ! We take the norm to be 1 (usually you do not want to use this,
     ! but it can be useful for debugging)
     !
     eqnorm = 1.d0
  elseif(itol.eq.1) then
     !
     ! The norm is simply the norm of the right-hand-side of the equation
     !
     eqnorm = bicg_norm(n,rhs,inorm)
  elseif(itol.eq.2) then
     !
     ! The norm is the norm of the solution of the approximate matrix equation
     !
     call msol(n,rhs,vec)
     eqnorm = bicg_norm(n,vec,inorm)
  else
     write(*,*) 'ERROR in BICG module: itol = ',itol,' is not implemented'
     stop
  endif
  !
  ! First compute the z and zb vectors. 
  !
  call msol(n,r,z)
  call mtrsol(n,r,zb)
  !
  ! Make initial values of p
  !
  p  = z
  pb = zb
  !
  ! Calculate one of the numbers used to compute alpha and beta
  !
  val1 = bicg_innprod(n,rb,z)
  !
  ! Do the iteration loop
  !
  do iter=1,itermax
     !
     ! Do a check
     !
     if(val1.eq.0.d0) then
        write(*,*) 'ERROR in bicg(): Zero rb*r value found...'
        stop
     endif
     !
     ! Compute alpha (see Eq. 2.7.32 of Num Recip in C)
     !
     call mx(n,p,vec)
     val2 = bicg_innprod(n,pb,vec)
     if(val2.ne.0.d0) then
        alpha = val1/val2
     else
        write(*,*) 'ERROR in bicg(): Zero pb*A*p value found...'
        stop
     endif
     !
     ! Now we can recompute the r and rb
     !
     r  = r - alpha * vec
     call mtrx(n,pb,vec)
     rb = rb - alpha * vec
     !
     ! We can also update the solution vector
     !
     sol = sol + alpha * p
     !
     ! Check if we have converged or not
     !
     err = bicg_norm(n,r,inorm) / eqnorm
     !
     ! Print message
     !
     write(*,*) 'Iteration ',iter,', Err = ',err
     !
     ! If converged, exit loop
     !
     if(err.le.tol) exit
     !
     ! Compute the z and zb
     !
     call msol(n,r,z)
     call mtrsol(n,rb,zb)
     !
     ! Now finish computing beta
     !
     val2 = bicg_innprod(n,rb,z)
     beta = val2 / val1
     !
     ! Now recompute p and pb
     !
     p  = z  + beta*p
     pb = zb + beta*pb
     !
     ! Swap val2 --> val2
     !
     val1 = val2
     !
  enddo
  !
  ! Check if we exceeded iter
  !
  if(iter.ge.itermax) then
     write(*,*) 'WARNING: No convergence reached yet. Error = ',err
  endif
  !
  ! Deallocate vectors r, rb, p, pb
  !
  deallocate(r,rb,p,pb,vec,z,zb)
  !
end subroutine solve_bicg

!-------------------------------------------------------------------------
!               COMPUTE THE INNER PRODUCT OF TWO VECTORS
!-------------------------------------------------------------------------
function bicg_innprod(n,a,b)
  implicit none
  integer :: n
  double precision :: a(n),b(n)
  integer :: i
  double precision :: bicg_innprod
  bicg_innprod = 0.d0
  do i=1,n
     bicg_innprod = bicg_innprod + a(i)*b(i)
  enddo
  return
end function bicg_innprod

!-------------------------------------------------------------------------
!                   COMPUTE THE NORM OF A VECTOR
!-------------------------------------------------------------------------
function bicg_norm(n,x,inorm)
  implicit none
  integer :: n,inorm
  double precision :: x(n)
  integer :: i
  double precision :: bicg_norm
  if(inorm.eq.1) then
     bicg_norm = 0.d0
     do i=1,n
        bicg_norm = bicg_norm + x(i)*x(i)
     enddo
  else
     bicg_norm = 0.d0
     do i=1,n
        if(abs(x(i)).gt.bicg_norm) then
           bicg_norm = abs(x(i))
        endif
     enddo
  endif
end function bicg_norm

end module bicg_module
